///////////////////////////////////////////////////////////////////////////////////////////////////
//
// PROJECT : ARIRANG family(ARIRANG-224,ARIRANG-256,ARIRANG-384,ARIRANG-512)
//
// DATE    : 2008.10.23
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FILE  : ARIRANG_7S64.c
//
// NOTES : Code of ARIRANG family
// 
//         Based on 64-bit platform (seven S-box version)
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <memory.h>
#include "Arirang_7S64.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Macro

#undef SS
#undef SSS

#define SS(x,n)			ROTL_DWORD(x,n)
#define SSS(x,n)		ROTL_QWORD(x,n)

#define f_mult2(x) (x&0x80 ? (0x1b^(x<<1)) : (x<<1))
#define ff_mult(a, b)	(a && b ? pow_tab[(log_tab[a] + log_tab[b]) % 255] : 0)
#define byte(x, n)		((BYTE)((x) >> (8 * n)))

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constant


static const DWORD K256[16] = {0x517cc1b7, 0x76517cc1, 0xbd76517c, 0x2dbd7651,
							   0x272dbd76, 0xcb272dbd, 0x90cb272d, 0x0a90cb27,
							   0xec0a90cb, 0x5bec0a90, 0x9a5bec0a, 0xe69a5bec,
							   0xb7e69a5b, 0xc1b7e69a, 0x7cc1b7e6, 0x517cc1b7};

static const QWORD K512[16] = {0x517cc1b727220a94, 0x2db6517cc1b72722, 0xe6952db6517cc1b7, 0x90cbe6952db6517c, 
								  0x7cca90cbe6952db6, 0xcb237cca90cbe695, 0x765ecb237cca90cb, 0xec01765ecb237cca, 
								  0xb7e9ec01765ecb23, 0xbd7db7e9ec01765e, 0x9a5fbd7db7e9ec01, 0x5be89a5fbd7db7e9, 
								  0x0a945be89a5fbd7d, 0x27220a945be89a5f, 0xc1b727220a945be8, 0x517cc1b727220a94};

BYTE	sbx1[256]; // S-box: S[i]
BYTE	sbx2[256]; // 2*S[i] GF(256)
BYTE	sbx3[256]; // 3*S[i] GF(256)
BYTE	sbx4[256]; // 4*S[i] GF(256)
BYTE	sbx8[256]; // 8*S[i] GF(256)
BYTE	sbx9[256]; // 9*S[i] GF(256)
BYTE	sbxA[256]; // 10*S[i] GF(256)



///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : gen_tabs
//
// DESCRIPTION    : Generate the s-box table and field multiplication table
//
// PARAMETERS     : void
//                  
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void gen_tabs(void)
{   

	#ifndef __GEN_TABS__
	#define __GEN_TABS__ 100
	BYTE i; 	
	BYTE p, q;

	/* log and power tables for GF(2**8) finite field with  */
	/* 0x011b as modular polynomial - the simplest primitive */
	/* root is 0x03, used here to generate the tables       */

	i = 0;p = 1;
	while(1)
	{
		sbx4[i] = (BYTE)p; //pow_tab ==> sbx4
		sbx8[p] = (BYTE)i; //log_tab ==> sbx8 
		p = p ^ (p << 1) ^ (p & 0x80 ? 0x01b : 0);
		if(i == 255) break;
		i++;
	}

	/* note that the affine byte transformation matrix in   */
	/* ARIRANG specification is in big endian format with  */
	/* bit 0 as the most significant bit. In the remainder  */
	/* of the specification the bits are numbered from the  */
	/* least significant end of a byte.                     */

	i = 0;
	while(1)    
	{   
		p = (i ? sbx4[255 - sbx8[i]] : 0); q = p; 
		q = (q >> 7) | (q << 1); p ^= q; 
		q = (q >> 7) | (q << 1); p ^= q; 
		q = (q >> 7) | (q << 1); p ^= q; 
		q = (q >> 7) | (q << 1); p ^= q ^ 0x63; 
		sbx1[i] = p;
		if(i == 0xff) break;
		i++;
	}
	i = 0;
	while(1)    
	{   
		sbx2[i] = f_mult2(sbx1[i]);		
		sbx4[i] = f_mult2(sbx2[i]);
		sbx8[i] = f_mult2(sbx4[i]);
		sbx3[i] = sbx1[i]^sbx2[i];	
		sbx9[i] = sbx1[i]^sbx8[i];
		sbxA[i] = sbx2[i]^sbx8[i];
		if(i == 0xff) break;
		i++;
	}
	#endif
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step256
//
// DESCRIPTION    : Step function of ARIRANG-224 and Arirang-256
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#define Step256(A, B, C, D, E, F, G, H, M1, M2){													\
	A ^= M1;				   E ^= M2;																\
	temp = ((sbx2[byte(A, 0)] ^ sbx3[byte(A, 1)] ^ sbx1[byte(A, 2)] ^ sbx1[byte(A, 3)])      ) ^	\
	       ((sbx1[byte(A, 0)] ^ sbx2[byte(A, 1)] ^ sbx3[byte(A, 2)] ^ sbx1[byte(A, 3)]) <<  8) ^	\
	       ((sbx1[byte(A, 0)] ^ sbx1[byte(A, 1)] ^ sbx2[byte(A, 2)] ^ sbx3[byte(A, 3)]) << 16) ^	\
	       ((sbx3[byte(A, 0)] ^ sbx1[byte(A, 1)] ^ sbx1[byte(A, 2)] ^ sbx2[byte(A, 3)]) << 24);		\
	B ^= temp;																						\
	C ^= SS(temp, 13);																				\
	D ^= SS(temp, 23);																				\
	temp = ((sbx2[byte(E, 0)] ^ sbx3[byte(E, 1)] ^ sbx1[byte(E, 2)] ^ sbx1[byte(E, 3)])      ) ^	\
		   ((sbx1[byte(E, 0)] ^ sbx2[byte(E, 1)] ^ sbx3[byte(E, 2)] ^ sbx1[byte(E, 3)]) <<  8) ^	\
		   ((sbx1[byte(E, 0)] ^ sbx1[byte(E, 1)] ^ sbx2[byte(E, 2)] ^ sbx3[byte(E, 3)]) << 16) ^	\
	       ((sbx3[byte(E, 0)] ^ sbx1[byte(E, 1)] ^ sbx1[byte(E, 2)] ^ sbx2[byte(E, 3)]) << 24);		\
	F ^= temp;																						\
	G ^= SS(temp, 29);																				\
	H ^= SS(temp, 7);																				\
}																									\

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step512
//
// DESCRIPTION    : Step function of ARIRANG-384 and Arirang-512
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
#define Step512(A, B, C, D, E, F, G, H, M1, M2){																														\
	A ^= M1;					                            E ^= M2;																									\
	temp    = ((QWORD)(sbx1[byte(A,0)] ^ sbx2[byte(A,1)] ^ sbxA[byte(A,2)] ^ sbx9[byte(A,3)] ^ sbx8[byte(A,4)] ^ sbx1[byte(A,5)] ^ sbx4[byte(A,6)] ^ sbx1[byte(A,7)])      ) ^	\
	          ((QWORD)(sbx1[byte(A,0)] ^ sbx1[byte(A,1)] ^ sbx2[byte(A,2)] ^ sbxA[byte(A,3)] ^ sbx9[byte(A,4)] ^ sbx8[byte(A,5)] ^ sbx1[byte(A,6)] ^ sbx4[byte(A,7)]) <<  8) ^	\
		      ((QWORD)(sbx4[byte(A,0)] ^ sbx1[byte(A,1)] ^ sbx1[byte(A,2)] ^ sbx2[byte(A,3)] ^ sbxA[byte(A,4)] ^ sbx9[byte(A,5)] ^ sbx8[byte(A,6)] ^ sbx1[byte(A,7)]) << 16) ^	\
			  ((QWORD)(sbx1[byte(A,0)] ^ sbx4[byte(A,1)] ^ sbx1[byte(A,2)] ^ sbx1[byte(A,3)] ^ sbx2[byte(A,4)] ^ sbxA[byte(A,5)] ^ sbx9[byte(A,6)] ^ sbx8[byte(A,7)]) << 24) ^	\
			  ((QWORD)(sbx8[byte(A,0)] ^ sbx1[byte(A,1)] ^ sbx4[byte(A,2)] ^ sbx1[byte(A,3)] ^ sbx1[byte(A,4)] ^ sbx2[byte(A,5)] ^ sbxA[byte(A,6)] ^ sbx9[byte(A,7)]) << 32) ^	\
			  ((QWORD)(sbx9[byte(A,0)] ^ sbx8[byte(A,1)] ^ sbx1[byte(A,2)] ^ sbx4[byte(A,3)] ^ sbx1[byte(A,4)] ^ sbx1[byte(A,5)] ^ sbx2[byte(A,6)] ^ sbxA[byte(A,7)]) << 40) ^	\
			  ((QWORD)(sbxA[byte(A,0)] ^ sbx9[byte(A,1)] ^ sbx8[byte(A,2)] ^ sbx1[byte(A,3)] ^ sbx4[byte(A,4)] ^ sbx1[byte(A,5)] ^ sbx1[byte(A,6)] ^ sbx2[byte(A,7)]) << 48) ^	\
			  ((QWORD)(sbx2[byte(A,0)] ^ sbxA[byte(A,1)] ^ sbx9[byte(A,2)] ^ sbx8[byte(A,3)] ^ sbx1[byte(A,4)] ^ sbx4[byte(A,5)] ^ sbx1[byte(A,6)] ^ sbx1[byte(A,7)]) << 56);	\
	B ^= temp;																																							\
	C ^= SSS(temp, 29);																																					\
	D ^= SSS(temp, 41);																																					\
	temp    = ((QWORD)(sbx1[byte(E,0)] ^ sbx2[byte(E,1)] ^ sbxA[byte(E,2)] ^ sbx9[byte(E,3)] ^ sbx8[byte(E,4)] ^ sbx1[byte(E,5)] ^ sbx4[byte(E,6)] ^ sbx1[byte(E,7)])      ) ^	\
		      ((QWORD)(sbx1[byte(E,0)] ^ sbx1[byte(E,1)] ^ sbx2[byte(E,2)] ^ sbxA[byte(E,3)] ^ sbx9[byte(E,4)] ^ sbx8[byte(E,5)] ^ sbx1[byte(E,6)] ^ sbx4[byte(E,7)]) <<  8) ^	\
			  ((QWORD)(sbx4[byte(E,0)] ^ sbx1[byte(E,1)] ^ sbx1[byte(E,2)] ^ sbx2[byte(E,3)] ^ sbxA[byte(E,4)] ^ sbx9[byte(E,5)] ^ sbx8[byte(E,6)] ^ sbx1[byte(E,7)]) << 16) ^	\
			  ((QWORD)(sbx1[byte(E,0)] ^ sbx4[byte(E,1)] ^ sbx1[byte(E,2)] ^ sbx1[byte(E,3)] ^ sbx2[byte(E,4)] ^ sbxA[byte(E,5)] ^ sbx9[byte(E,6)] ^ sbx8[byte(E,7)]) << 24) ^	\
			  ((QWORD)(sbx8[byte(E,0)] ^ sbx1[byte(E,1)] ^ sbx4[byte(E,2)] ^ sbx1[byte(E,3)] ^ sbx1[byte(E,4)] ^ sbx2[byte(E,5)] ^ sbxA[byte(E,6)] ^ sbx9[byte(E,7)]) << 32) ^	\
			  ((QWORD)(sbx9[byte(E,0)] ^ sbx8[byte(E,1)] ^ sbx1[byte(E,2)] ^ sbx4[byte(E,3)] ^ sbx1[byte(E,4)] ^ sbx1[byte(E,5)] ^ sbx2[byte(E,6)] ^ sbxA[byte(E,7)]) << 40) ^	\
			  ((QWORD)(sbxA[byte(E,0)] ^ sbx9[byte(E,1)] ^ sbx8[byte(E,2)] ^ sbx1[byte(E,3)] ^ sbx4[byte(E,4)] ^ sbx1[byte(E,5)] ^ sbx1[byte(E,6)] ^ sbx2[byte(E,7)]) << 48) ^	\
			  ((QWORD)(sbx2[byte(E,0)] ^ sbxA[byte(E,1)] ^ sbx9[byte(E,2)] ^ sbx8[byte(E,3)] ^ sbx1[byte(E,4)] ^ sbx4[byte(E,5)] ^ sbx1[byte(E,6)] ^ sbx1[byte(E,7)]) << 56);	\
	F ^= temp;																																							\
	G ^= SSS(temp, 53);																																					\
	H ^= SSS(temp, 13);																																					\
}																																										\

#define COUNT512() {state->counter[0]++; if(state->counter[0] == 0)	state->counter[1]++;}
#define COUNT256() {state->counter[0]++;}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Arirang_Compression256
//
// DESCRIPTION    : Counter addition and compression function of Arirang-224 and ARIRANG-256
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Arirang_Compression256(hashState *state)
{
	DWORD	R[8], W[32], temp;
	int i;

	#if defined(BIG_ENDIAN)
		#define GetData(x) x
	#else
		#define GetData(x) ENDIAN_REVERSE_DWORD(x)
	#endif

	// Message Scheduling
	for (i = 0; i < 16; i++)
		W[i] = GetData(((DWORD*)state->block)[i]);

	W[16] = SS((W[ 9] ^ W[11] ^ W[13] ^ W[15] ^ K256[ 0]),  5);
	W[17] = SS((W[ 8] ^ W[10] ^ W[12] ^ W[14] ^ K256[ 1]), 11);
	W[18] = SS((W[ 1] ^ W[ 3] ^ W[ 5] ^ W[ 7] ^ K256[ 2]), 19);
	W[19] = SS((W[ 0] ^ W[ 2] ^ W[ 4] ^ W[ 6] ^ K256[ 3]), 31);

	W[20] = SS((W[14] ^ W[ 4] ^ W[10] ^ W[ 0] ^ K256[ 4]),  5);
	W[21] = SS((W[11] ^ W[ 1] ^ W[ 7] ^ W[13] ^ K256[ 5]), 11);
	W[22] = SS((W[ 6] ^ W[12] ^ W[ 2] ^ W[ 8] ^ K256[ 6]), 19);
	W[23] = SS((W[ 3] ^ W[ 9] ^ W[15] ^ W[ 5] ^ K256[ 7]), 31);

	W[24] = SS((W[13] ^ W[15] ^ W[ 1] ^ W[ 3] ^ K256[ 8]),  5);
	W[25] = SS((W[ 4] ^ W[ 6] ^ W[ 8] ^ W[10] ^ K256[ 9]), 11);
	W[26] = SS((W[ 5] ^ W[ 7] ^ W[ 9] ^ W[11] ^ K256[10]), 19);
	W[27] = SS((W[12] ^ W[14] ^ W[ 0] ^ W[ 2] ^ K256[11]), 31);

	W[28] = SS((W[10] ^ W[ 0] ^ W[ 6] ^ W[12] ^ K256[12]),  5);
	W[29] = SS((W[15] ^ W[ 5] ^ W[11] ^ W[ 1] ^ K256[13]), 11);
	W[30] = SS((W[ 2] ^ W[ 8] ^ W[14] ^ W[ 4] ^ K256[14]), 19);
	W[31] = SS((W[ 7] ^ W[13] ^ W[ 3] ^ W[ 9] ^ K256[15]), 31);

	// Counter Addition
	state->workingvar[0] ^= ((DWORD*)state->counter)[1];
	state->workingvar[4] ^= ((DWORD*)state->counter)[0];

	// Initialize Register
	R[0] = state->workingvar[0];
	R[1] = state->workingvar[1];
	R[2] = state->workingvar[2];
	R[3] = state->workingvar[3];
	R[4] = state->workingvar[4];
	R[5] = state->workingvar[5];
	R[6] = state->workingvar[6];
	R[7] = state->workingvar[7];
	
	// 1 Round
	Step256(R[0], R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[16], W[17]);
	Step256(R[7], R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 0], W[ 1]);
	Step256(R[6], R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 2], W[ 3]);
	Step256(R[5], R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 4], W[ 5]);
	Step256(R[4], R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 6], W[ 7]);

	Step256(R[3], R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[18], W[19]);
	Step256(R[2], R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[ 8], W[ 9]);
	Step256(R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[10], W[11]);
	Step256(R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[12], W[13]);
	Step256(R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[14], W[15]);

	// 2 Round
	Step256(R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[20], W[21]);
	Step256(R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 3], W[ 6]);
	Step256(R[4], R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 9], W[12]);
	Step256(R[3], R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[15], W[ 2]);
	Step256(R[2], R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[ 5], W[ 8]);

	Step256(R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[22], W[23]);
	Step256(R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[11], W[14]);
	Step256(R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 1], W[ 4]);
	Step256(R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 7], W[10]);
	Step256(R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[13], W[ 0]);

	// Feedforward_1
	R[0] ^= state->workingvar[4];
	R[1] ^= state->workingvar[5];
	R[2] ^= state->workingvar[6];
	R[3] ^= state->workingvar[7];
	R[4] ^= state->workingvar[0];
	R[5] ^= state->workingvar[1];
	R[6] ^= state->workingvar[2];
	R[7] ^= state->workingvar[3];

	// 3 Round
	Step256(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[24], W[25]);
	Step256(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[12], W[ 5]);
	Step256(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[14], W[ 7]);
	Step256(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 0], W[ 9]);
	Step256(R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[ 2], W[11]);

	Step256(R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[26], W[27]);
	Step256(R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 4], W[13]);
	Step256(R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 6], W[15]);
	Step256(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 8], W[ 1]);
	Step256(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[10], W[ 3]);

	// 4 Round
	Step256(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[28], W[29]);
	Step256(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 7], W[ 2]);
	Step256(R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[13], W[ 8]);
	Step256(R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 3], W[14]);
	Step256(R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 9], W[ 4]);

	Step256(R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[30], W[31]);
	Step256(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[15], W[10]);
	Step256(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[ 5], W[ 0]);
	Step256(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[11], W[ 6]);
	Step256(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 1], W[12]);

	// Feedforward_2
	state->workingvar[0] ^= R[0];
	state->workingvar[1] ^= R[1];
	state->workingvar[2] ^= R[2];
	state->workingvar[3] ^= R[3];
	state->workingvar[4] ^= R[4];
	state->workingvar[5] ^= R[5];
	state->workingvar[6] ^= R[6];
	state->workingvar[7] ^= R[7];

	COUNT256();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : ARIRANG_Compression512
//
// DESCRIPTION    : Counter Addition and compression function of ARIRANG-384 and ARIRANG-512
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void Arirang_Compression512(hashState *state)
{
	QWORD	R[8], W[32], temp;
	int i;

	#if defined(BIG_ENDIAN)
		#define GetData(x) x
	#else
		#define GetData(x) ENDIAN_REVERSE_DWORD(x)
	#endif

	// Message Scheduling
	for (i = 0; i < 16; i++)
		W[i] = (QWORD)(GetData(((DWORD*)state->block)[2*i+1])) | ((QWORD)(GetData(((DWORD*)state->block)[2*i])) << 32); 
											
	W[16] = SSS((W[ 9] ^ W[11] ^ W[13] ^ W[15] ^ K512[ 0]), 11);
	W[17] = SSS((W[ 8] ^ W[10] ^ W[12] ^ W[14] ^ K512[ 1]), 23);
	W[18] = SSS((W[ 1] ^ W[ 3] ^ W[ 5] ^ W[ 7] ^ K512[ 2]), 37);
	W[19] = SSS((W[ 0] ^ W[ 2] ^ W[ 4] ^ W[ 6] ^ K512[ 3]), 59);

	W[20] = SSS((W[14] ^ W[ 4] ^ W[10] ^ W[ 0] ^ K512[ 4]), 11);
	W[21] = SSS((W[11] ^ W[ 1] ^ W[ 7] ^ W[13] ^ K512[ 5]), 23);
	W[22] = SSS((W[ 6] ^ W[12] ^ W[ 2] ^ W[ 8] ^ K512[ 6]), 37);
	W[23] = SSS((W[ 3] ^ W[ 9] ^ W[15] ^ W[ 5] ^ K512[ 7]), 59);

	W[24] = SSS((W[13] ^ W[15] ^ W[ 1] ^ W[ 3] ^ K512[ 8]), 11);
	W[25] = SSS((W[ 4] ^ W[ 6] ^ W[ 8] ^ W[10] ^ K512[ 9]), 23);
	W[26] = SSS((W[ 5] ^ W[ 7] ^ W[ 9] ^ W[11] ^ K512[10]), 37);
	W[27] = SSS((W[12] ^ W[14] ^ W[ 0] ^ W[ 2] ^ K512[11]), 59);

	W[28] = SSS((W[10] ^ W[ 0] ^ W[ 6] ^ W[12] ^ K512[12]), 11);
	W[29] = SSS((W[15] ^ W[ 5] ^ W[11] ^ W[ 1] ^ K512[13]), 23);
	W[30] = SSS((W[ 2] ^ W[ 8] ^ W[14] ^ W[ 4] ^ K512[14]), 37);
	W[31] = SSS((W[ 7] ^ W[13] ^ W[ 3] ^ W[ 9] ^ K512[15]), 59);
																												

	// Counter Addition
	state->workingvar[0] ^= state->counter[1];
	state->workingvar[4] ^= state->counter[0];

	// Initialize Register
	R[0] = state->workingvar[0];
	R[1] = state->workingvar[1];
	R[2] = state->workingvar[2];
	R[3] = state->workingvar[3];
	R[4] = state->workingvar[4];
	R[5] = state->workingvar[5];
	R[6] = state->workingvar[6];
	R[7] = state->workingvar[7];
	
	// 1 Round
	Step512(R[0], R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[16], W[17]);
	Step512(R[7], R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 0], W[ 1]);
	Step512(R[6], R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 2], W[ 3]);
	Step512(R[5], R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 4], W[ 5]);
	Step512(R[4], R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 6], W[ 7]);

	Step512(R[3], R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[18], W[19]);
	Step512(R[2], R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[ 8], W[ 9]);
	Step512(R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[10], W[11]);
	Step512(R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[12], W[13]);
	Step512(R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[14], W[15]);

	// 2 Round
	Step512(R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[20], W[21]);
	Step512(R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 3], W[ 6]);
	Step512(R[4], R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 9], W[12]);
	Step512(R[3], R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[15], W[ 2]);
	Step512(R[2], R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[ 5], W[ 8]);

	Step512(R[1], R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[22], W[23]);
	Step512(R[0], R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[11], W[14]);
	Step512(R[7], R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 1], W[ 4]);
	Step512(R[6], R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 7], W[10]);
	Step512(R[5], R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[13], W[ 0]);

	// Feedforward_1
	R[0] ^= state->workingvar[4];
	R[1] ^= state->workingvar[5];
	R[2] ^= state->workingvar[6];
	R[3] ^= state->workingvar[7];
	R[4] ^= state->workingvar[0];
	R[5] ^= state->workingvar[1];
	R[6] ^= state->workingvar[2];
	R[7] ^= state->workingvar[3];

	// 3 Round
	Step512(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[24], W[25]);
	Step512(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[12], W[ 5]);
	Step512(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[14], W[ 7]);
	Step512(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 0], W[ 9]);
	Step512(R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[ 2], W[11]);

	Step512(R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[26], W[27]);
	Step512(R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 4], W[13]);
	Step512(R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[ 6], W[15]);
	Step512(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[ 8], W[ 1]);
	Step512(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[10], W[ 3]);

	// 4 Round
	Step512(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[28], W[29]);
	Step512(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 7], W[ 2]);
	Step512(R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7], W[13], W[ 8]);
	Step512(R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6], W[ 3], W[14]);
	Step512(R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4] ,R[5], W[ 9], W[ 4]);

	Step512(R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3] ,R[4], W[30], W[31]);
	Step512(R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2] ,R[3], W[15], W[10]);
	Step512(R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1] ,R[2], W[ 5], W[ 0]);
	Step512(R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0] ,R[1], W[11], W[ 6]);
	Step512(R[1] ,R[2] ,R[3] ,R[4] ,R[5] ,R[6] ,R[7] ,R[0], W[ 1], W[12]);

	// Feedforward_2
	state->workingvar[0] ^= R[0];
	state->workingvar[1] ^= R[1];
	state->workingvar[2] ^= R[2];
	state->workingvar[3] ^= R[3];
	state->workingvar[4] ^= R[4];
	state->workingvar[5] ^= R[5];
	state->workingvar[6] ^= R[6];
	state->workingvar[7] ^= R[7];
	
	COUNT512();														
}

